# -*- coding: utf-8 -*-

"""
/***************************************************************************
 ForstProcessing
                                 A QGIS plugin
 ForstProcessing
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2018-11-13
        copyright            : (C) 2018 by CHL
        email                : christina_ludwig@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = 'CHL'
__date__ = '2018-11-13'
__copyright__ = '(C) 2018 by CHL'

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = '$Format:%H$'

import os
from PyQt5.QtCore import QCoreApplication, QVariant, Qt
from PyQt5.QtWidgets import QMessageBox
from qgis.PyQt.QtGui import QIcon, QFont
from qgis.core import (QgsProcessing, QgsVectorLayer, QgsField, QgsSettings, QgsProject, QgsWkbTypes,
                       QgsCoordinateReferenceSystem, QgsPointXY, QgsPoint,
                       QgsProcessingAlgorithm,
                       QgsProcessingParameterString, QgsDistanceArea,
                       QgsProcessingParameterFeatureSink, QgsFeature, QgsGeometry, QgsPalLayerSettings, QgsTextFormat,
                       QgsVectorLayerSimpleLabeling)
from qgis.gui import QgsMessageBar
from .calculategeometry_dialog import CalculateGeometryDialog
from .help_dialog import HelpDialog
from processing.gui.MessageBarProgress import MessageBarProgress
from processing.tools import dataobjects
from qgis.utils import iface

from ..resources import *
from .utils import *

# todo Check order of columns in attribute table

class CalculateGeometry(QgsProcessingAlgorithm):
    """
    This is an example algorithm that takes a vector layer and
    creates a new identical one.

    It is meant to be used as an example of how to create your own
    algorithms and explain methods and variables used to do it. An
    algorithm like this will be available in all elements, and there
    is not need for additional work.

    All Processing algorithms should extend the QgsProcessingAlgorithm
    class.
    """

    # Constants used to refer to parameters and outputs. They will be
    # used when calling the algorithm from another algorithm, or when
    # calling from the QGIS console.

    OUTPUT = ""
    parent = None

    def initAlgorithm(self, config):
        """
        Here we define the inputs and output of the algorithm, along
        with some other properties.
        """

        # We add a feature sink in which to store our processed features (this
        # usually takes the form of a newly created vector layer when the
        # algorithm is run in QGIS).
        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr('Output layer')
            )
        )

        self.canvas = iface.mapCanvas()

        self.dlg = CalculateGeometryDialog()
        self.help = HelpDialog()

        # Connect signals and slots
        self.dlg.buttonBox.rejected.connect(self.reject)
        self.dlg.buttonBox.accepted.connect(self.check_input)
        self.dlg.btn_help.clicked.connect(self.help.show)

    def run(self):
        self.processAlgorithm()

    def processAlgorithm(self):
        """
        Here is where the processing itself takes place.
        """

        print("Start processing")
        self.layer = self.canvas.currentLayer()
        self.export_geometry_info(self.layer)
        iface.showAttributeTable(self.layer)

        return {self.OUTPUT: "dest_id"}

    def name(self):
        """
        Returns the algorithm name, used for identifying the algorithm. This
        string should be fixed for the algorithm, and must not be localised.
        The name should be unique within each provider. Names should contain
        lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return '2 - Werteberechnung Attributtabelle'

    def displayName(self):
        """
        Returns the translated algorithm name, which should be used for any
        user-visible display of the algorithm name.
        """
        return self.tr(self.name())

    def tr(self, string):
        return QCoreApplication.translate('Processing', string)

    def createInstance(self):
        return CalculateGeometry()

    def createCustomParametersWidget(self, parent):
        self.parent = parent
        return self.dlg

    def svgIconPath(self):
        return os.path.abspath("./icons/calculate_geometry.png")

    def icon(self):
        return QIcon(":/plugins/icons/calculate_geometry.png")

    def reject(self):
        self.dlg.close()

    def check_input(self):
        """
        Check the input parameters

        """
        layer = self.canvas.currentLayer()

        if layer is None:
            # iface.messageBar().pushWarning("Fehler",
            #                               u"Bitte wählen Sie einen Vektorlayer als Eingabelayer aus.")
            QMessageBox.warning(iface.mainWindow(),
                                "Warnung",
                                u"Bitte wählen Sie eine Vektorlayer aus.")
            return 1
        elif not isinstance(layer, QgsVectorLayer):
            # iface.messageBar().pushWarning("Fehler",
            #                               u"Der Eingabelayer muss ein Vektorlayer sein.")
            QMessageBox.warning(iface.mainWindow(),
                                "Warnung",
                                u"Der ausgewählte Layer ist kein Vektorlayer.")
            return 1
        elif layer.isEditable():
            # iface.messageBar().pushWarning("Fehler",
            #                               u"Der ausgewählte Layer ist im Editiermodus. Für die neu digitalisierten "
            #                               u"Features wird die Geometrie daher nicht berechnet.\n\nSpeichern Sie Ihre"
            #                               u" Änderungen und beenden Sie den Editiermodus, um die Geometrie aller ")
            QMessageBox.information(iface.mainWindow(),
                                    "Warnung",
                                    u"Der ausgewählte Layer ist im Editiermodus. Für die neu digitalisierten Features "
                                    u"wird die Geometrie daher nicht berechnet. \n\nSpeichern Sie Ihre Änderungen und "
                                    u"beenden Sie den Editiermodus um die Geometrien für alle Features zu berechnen.")
            return 1
        else:
            self.run()
            return 0

    def measure_point(self, geom):
        """
        Return coordinates of point geomtry
        :param geom:
        :return:
        """
        if geom.wkbType() in (QgsWkbTypes.MultiPoint, QgsWkbTypes.MultiPoint25D):
            point_geom = geom.asMultiPoint()
            x = point_geom[0].x()
            y = point_geom[0].y()
        else:
            point_geom = geom.asPoint()
            x = point_geom.x()
            y = point_geom.y()
        return x, y

    def measure_line(self, geom, calculator):
        #QgsDistanceArea().setSourceCrs(self.layer.crs())
        length = calculator.measureLength(geom)
        return length, length

    def measure_polygon(self, geom, calculator):
        #QgsDistanceArea().setSourceCrs(self.layer.crs())
        area = calculator.measureArea(geom) / 10000.
        perimeter = calculator.measurePerimeter(geom)
        return area, perimeter

    def find_or_create_fields(self, layer):
        fields = layer.fields()
        geom_type = layer.geometryType()
        field_name_list = [field.name().lower() for field in fields.toList()]
        new_fields = []
        n_fields = len(field_name_list)

        # POLYGON
        if geom_type == QgsWkbTypes.PolygonGeometry:
            # Add HEKTAR column
            found_hektar, index1 = self.field_exists(field_name_list, "HEKTAR")
            if not found_hektar:
                field = QgsField("HEKTAR", QVariant.Double, "double precision", 21, 4, "HEKTAR")
                index1 = n_fields
                new_fields.append(field)

            # Add METER column
            found_meter, index2 = self.field_exists(field_name_list, "METERS")
            if not found_meter:
                found_meter, index2 = self.field_exists(field_name_list, "METER")
            if not found_meter:
                field = QgsField("METER", QVariant.Double, "double precision", 21, 0, "METER")
                if not found_hektar:
                    index2 = n_fields + 1
                else:
                    index2 = n_fields
                new_fields.append(field)

            layer.dataProvider().addAttributes(new_fields)
            layer.updateFields()
            return index1, index2

        # LINE
        elif geom_type == QgsWkbTypes.LineGeometry:

            found_METERS, index1 = self.field_exists(field_name_list, "METERS")
            if not found_METERS:
                (found_METERS, index1) = self.field_exists(field_name_list, "METER")

            if not found_METERS:
                field = QgsField("METER", QVariant.Double, "double precision", 21, 0, "METER")
                index1 = n_fields
                new_fields.append(field)

            layer.dataProvider().addAttributes(new_fields)
            layer.updateFields()
            return index1, index1

        # POINT
        else:
            found_RECHTS, index1 = self.field_exists(field_name_list, "RECHTSWERT")
            if not found_RECHTS:
                field = QgsField("RECHTSWERT", QVariant.Double, "double precision", 21, 6, "RECHTSWERT")
                index1 = n_fields
                new_fields.append(field)

            found_HOCHWERT, index2 = self.field_exists(field_name_list, "HOCHWERT")
            if not found_HOCHWERT:
                field = QgsField("HOCHWERT", QVariant.Double, "double precision", 21, 6, "HOCHWERT")
                if not found_RECHTS:
                    index2 = n_fields + 1
                else:
                    index2 = n_fields
                new_fields.append(field)

            layer.dataProvider().addAttributes(new_fields)
            layer.updateFields()
            return index1, index2

    def field_exists(self, field_names, e):
        e = e.lower()
        for idx, fname in enumerate(field_names):
            if fname.lower() == e:
                return True, idx
        return False, len(field_names) - 1

    def export_geometry_info(self, layer):

        # update existing file
        index1, index2 = self.find_or_create_fields(layer)

        if self.dlg.ckbSelectedFeatures.checkState() == Qt.Checked:
            features = layer.getSelectedFeatures()
        else:
            features = layer.getFeatures()

        # Check CRS
        calculator = QgsDistanceArea()
        calculator.setSourceCrs(layer.crs(), QgsProject().instance().transformContext())

        for feat in features:
            geom = feat.geometry()
            # inGeom.transform( coordTransform )
            if QgsWkbTypes.wkbDimensions(geom.wkbType()) == 0:
                attr1, attr2 = self.measure_point(geom)
            elif QgsWkbTypes.wkbDimensions(geom.wkbType()) == 1:
                attr1, attr2 = self.measure_line(geom, calculator)
            else:
                attr1, attr2 = self.measure_polygon(geom, calculator)

            changeMap = {}
            changeAttributeMap = {index1: attr1,
                                  index2: attr2}
            changeMap[feat.id()] = changeAttributeMap
            layer.dataProvider().changeAttributeValues(changeMap)

        return True
