# -*- coding: utf-8 -*-

"""
/***************************************************************************
 ForstProcessing
                                 A QGIS plugin
 ForstProcessing
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2018-11-13
        copyright            : (C) 2018 by CHL
        email                : christina_ludwig@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = 'CHL'
__date__ = '2018-11-13'
__copyright__ = '(C) 2018 by CHL'

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = '$Format:%H$'

import os
from PyQt5.QtCore import QCoreApplication, QVariant, QUrl, QUrlQuery, Qt, QSettings
from PyQt5.QtWidgets import QMessageBox
from qgis.PyQt.QtGui import QIcon, QFont
from qgis.PyQt.QtNetwork import QNetworkRequest, QNetworkAccessManager
from qgis.core import (QgsProcessing, QgsVectorLayer, QgsField, QgsSettings, QgsProject, QgsApplication,
                       QgsCoordinateReferenceSystem, QgsPointXY, QgsPoint, QgsRectangle,
                       QgsProcessingAlgorithm, QgsRasterLayer, QgsNetworkAccessManager, QgsMessageLog,
                       QgsProcessingParameterString,
                       QgsProcessingParameterFeatureSink, QgsFeature, QgsGeometry, QgsPalLayerSettings, QgsTextFormat,
                       QgsVectorLayerSimpleLabeling)
from .address_lookup_dialog import addressLookupDialog
from processing.gui.MessageBarProgress import MessageBarProgress
from processing.tools import dataobjects
from qgis.utils import iface
import urllib3
import certifi
import ssl
import os
import sys
import subprocess
import json
import ogr
import requests
import urllib.parse

from ..resources import *
from .utils import *


class AddressLookup(QgsProcessingAlgorithm):
    """
    This is an example algorithm that takes a vector layer and
    creates a new identical one.

    It is meant to be used as an example of how to create your own
    algorithms and explain methods and variables used to do it. An
    algorithm like this will be available in all elements, and there
    is not need for additional work.

    All Processing algorithms should extend the QgsProcessingAlgorithm
    class.
    """

    # Constants used to refer to parameters and outputs. They will be
    # used when calling the algorithm from another algorithm, or when
    # calling from the QGIS console.

    query = ""
    OUTPUT = ""
    parent = None
    canvas = iface.mapCanvas()

    def initAlgorithm(self, config=None):
        """
        Here we define the inputs and output of the algorithm, along
        with some other properties.
        """

        self.parent = None
        self.layerid = ''
        self.srs = QgsCoordinateReferenceSystem("EPSG:32633")
        self.defaultPoint = pointTo3857(QgsPointXY(407719.7, 5279324.9), self.srs)

        if QgsSettings().value("addressLookup/mode") is None:
            QgsSettings().setValue("addressLookup/mode", 1)

        # We add the input vector features source. It can have any kind of
        # geometry.
        self.addParameter(
            QgsProcessingParameterString(
                self.query,
                'Adresse'
            )
        )

        # We add a feature sink in which to store our processed features (this
        # usually takes the form of a newly created vector layer when the
        # algorithm is run in QGIS).
        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr('Output layer')
            )
        )

        self.dlg = addressLookupDialog()
        self.dlg.buttonBox.rejected.connect(self.reject)
        self.dlg.buttonBox.accepted.connect(self.run)

    def show(self):
        self.dlg.exec_()

    def run(self):
        self.processAlgorithm()

    def get_input(self):
        if self.dlg.RADIO_1.isChecked():
            QgsSettings().setValue("addressLookup/mode", 'run_procedure')
        elif self.dlg.RADIO_2.isChecked():
            QgsSettings().setValue("addressLookup/mode", 'suppress_procedure')
        else:
            QgsSettings().setValue("addressLookup/mode", 'load_basemap')

        self.query = self.dlg.ADDRESS.text()

    def add_basemap(self):
        raster = QgsRasterLayer("type=xyz&url=http://ecn.t3.tiles.virtualearth.net/tiles/a{q}.jpeg?g=1", "bing", "wms")
        QgsProject().instance().addMapLayer(raster)
        QCoreApplication.processEvents()


    def process_point(self, place, point):
        """
        Transforms the point and save
        """
        # lon lat and transform
        destCRS = QgsCoordinateReferenceSystem("EPSG:32633")
        point = QgsPointXY(point[1], point[0])
        point_epsg32633 = pointFromWGS84(point, destCRS)
        # Save point
        self.save_point(point_epsg32633, str(place))
        return point_epsg32633

    def set_labels(self, layer, fieldname):
        layer_settings = QgsPalLayerSettings()
        text_format = QgsTextFormat()

        text_format.setFont(QFont("Arial", 12))
        text_format.setSize(12)
        layer_settings.setFormat(text_format)
        layer_settings.fieldName = fieldname
        layer_settings.placement = 4
        layer_settings.enabled = True
        layer_settings = QgsVectorLayerSimpleLabeling(layer_settings)
        layer.setLabelsEnabled(True)
        layer.setLabeling(layer_settings)
        layer.triggerRepaint()
        return layer

    def save_point(self, point, address):
        # create and add the point layer
        # create layer with same CRS as map canvas
        layer = QgsVectorLayer("Point?crs=EPSG:32633", "Ergebnis der Adresssuche", "memory")
        provider = layer.dataProvider()

        # Add new field
        provider.addAttributes([QgsField("Adresse", QVariant.String)])
        layer.updateFields()

        # Add a feature
        feat = QgsFeature(layer.fields())
        feat.setGeometry(QgsGeometry.fromPointXY(point))
        feat.setAttribute(0, address)
        layer.dataProvider().addFeatures([feat])
        layer.updateExtents()

        # Load to canvas and set labels
        QgsProject.instance().addMapLayer(layer)
        self.layerid = layer.id()
        self.set_labels(layer, "Adresse")
        QCoreApplication.processEvents()

    def zoomToPoint(self, point, scale):
        self.canvas.setCenter(point)
        self.canvas.zoomScale(scale)

    def zoomStandard(self):
        print("zoom in ")
        default_point = QgsPointXY(13.770952, 47.660733)
        point_mapCRS = pointFromWGS84(default_point, QgsProject().instance().crs())
        self.zoomToPoint(point_mapCRS, 2200000)

    def processAlgorithm(self):
        """
        Here is where the processing itself takes place.
        """

        print("Start processing")
        self.get_input()
        self.geocode()

        return {self.OUTPUT: "dest_id"}

    def geocode(self):
        try:
            self.send_query(self.query)
        except Exception as e:
            for m in e.args:
                QgsMessageLog.logMessage(m, 'Extensions')
            pass

    def send_query(self, query):
        print("got query")
        query = urllib.parse.quote(query)
        params = {'addressdetails': '0',
                  "polygon_geojson": "1",
                  "format": "json"}
        uri = 'https://nominatim.openstreetmap.org/search/'
        reply = requests.get(uri+query, params=params)
        try:
            r = json.loads(reply.text)
            geom = ogr.CreateGeometryFromJson(json.dumps(r[0]["geojson"])).Centroid()
            geom = geom.GetPoint(0)
            address = r[0]['display_name']
            if address is not None:
                self.add_basemap()
                point = (geom[1], geom[0])
                point_mapCRS = pointFromWGS84(QgsPointXY(point[1], point[0]), QgsProject().instance().crs())
                self.process_point(address, point)
                self.zoomToPoint(point_mapCRS, 5000)
        except Exception as e:
            print(e)

    def name(self):
        """
        Returns the algorithm name, used for identifying the algorithm. This
        string should be fixed for the algorithm, and must not be localised.
        The name should be unique within each provider. Names should contain
        lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return '6 - Adress-Suche'

    def displayName(self):
        """
        Returns the translated algorithm name, which should be used for any
        user-visible display of the algorithm name.
        """
        return self.tr(self.name())

    def tr(self, string):
        return QCoreApplication.translate('Processing', string)

    def createInstance(self):
        return AddressLookup()

    def createCustomParametersWidget(self, parent):
        self.parent = parent
        return self.dlg

    def svgIconPath(self):
        return os.path.abspath("./icons/address_lookup.png")

    def icon(self):
        return QIcon(":/plugins/icons/address_lookup.png")

    def reject(self):
        self.dlg.close()

